import os
import tarfile
import string
from os import path
from AccessControl import ClassSecurityInfo
from Products.CMFCore.utils import UniqueObject, getToolByName
from OFS.PropertyManager import PropertyManager
from OFS.SimpleItem import SimpleItem
from Products.BTreeFolder2.CMFBTreeFolder import CMFBTreeFolder
from Globals import InitializeClass
from time import time
from datetime import datetime
from ZODB.PersistentMapping import PersistentMapping

from Products.SQLExporter.content_exporter import Exporter
from Products.SQLExporter.config import database_path

live_exporters = {}

def elapsed_time(start,end):
    if start and end:
        delta = end - start
    elif start:
        delta = datetime.now() - start
    else:
        return '--'
    t = ''
    if delta.days:
        t += '%i days, ' % delta.days
    hrs = delta.seconds/(3600)
    t += '%s:' % string.zfill(hrs,2)
    mins = (delta.seconds - hrs*3600)/60
    t += '%s:' % string.zfill(mins,2)
    secs = (delta.seconds - hrs*3600 - mins*60)
    t += '%s:%s' % (string.zfill(secs,2),delta.microseconds)
    return t

class SQLExporterTool(PropertyManager, UniqueObject, SimpleItem):
    """ SQLExporterTool  .... """
    id = 'sql_exporter_tool'
    meta_type= 'SQL Exporter Tool'
    plone_tool = 1
    excluded_ids =  ('portal_memberdata',)
    excluded_meta_types =  ()

    _properties = (
        {'id':'excluded_ids', 'type': 'lines', 'mode':'w'},
        {'id':'excluded_meta_types', 'type': 'lines', 'mode':'w'},
    )

    manage_options = (

        ({'label' : 'Overview',
          'action' : 'overview'
          },)
        + PropertyManager.manage_options
        + SimpleItem.manage_options
    )


    security = ClassSecurityInfo()

    security.declareProtected('Manage Portal','manage_afterAdd')
    def manage_afterAdd(self, item, container):
        """post creation (or post renaming) actions"""
        result = super(SQLExporterTool,self).manage_afterAdd(item,container)
        self.old_exporters = PersistentMapping()
        return result


    security.declareProtected('Manage Portal','create_exporter')
    def create_exporter(self,from_path):
        xptr_id = datetime.now().strftime('%F:%H:%M:%S')
        xptr = Exporter(id=xptr_id,
                        from_path=from_path,
                        database_path=self.get_database_path(),
                        excluded_ids=self.getProperty('excluded_ids'),
                        excluded_meta_types=self.getProperty('excluded_meta_types'),
                       )
        live_exporters[xptr_id] = xptr
        return xptr_id

    security.declareProtected('Manage Portal','run_exporter')
    def run_exporter(self,REQUEST=None,RESPOSE=None):
        """starts exporter if available"""
        xptr_id = REQUEST.form.get('xptr_id','')
        if not xptr_id:
            return 'empty exporter id'
        if xptr_id not in live_exporters:
            return 'exporter id %s not alive' % xptr_id
        xptr = live_exporters[xptr_id]
        if xptr.running:
            return 'exporter already running'
        try:
            from_object = self.portal_url.getPortalObject().restrictedTraverse(xptr.from_path)
        except Exception, e:
            return 'error fetching export starting point object: %s' % e
        result = xptr.run(from_object)
        live_exporters.pop(xptr_id)
        self.old_exporters[xptr_id] = xptr
        print 'export done'
        return result

    security.declareProtected('Manage Portal','get_database_path')
    def get_database_path(self):
        if database_path is not None:
            return database_path
        ihome = os.getenv('INSTANCE_HOME')
        sql_xptr_home = os.path.join(ihome,'var','sql_exporter')
        if not os.path.exists(sql_xptr_home):
            os.makedirs(sql_xptr_home)
        return sql_xptr_home

    security.declareProtected('Manage Portal','results_sql_exporter_xml')
    def results_sql_exporter_xml(self,REQUEST=None,RESPOSE=None):
        """xml response call from ajax"""
        REQUEST.response.setHeader('content-type','text/xml')
        any_running = False
        live_items = '<items>'
        for xptr_id,xptr in live_exporters.items():
            any_running = any_running or xptr.running
            live_items += self._xptr_info(xptr_id,xml=True)

        for xptr_id,xptr in self.old_exporters.items():
            any_running = any_running or xptr.running
            live_items += self._xptr_info(xptr_id,xml=True)
        live_items += '</items>'

        return '<results working="%s">%s</results>' % (str(any_running).lower(),live_items)

    security.declareProtected('Manage Portal','get_old_exporters')
    def get_live_exporters(self):
        keys = live_exporters.keys()
        keys.sort()
        keys.reverse()
        return keys

    security.declareProtected('Manage Portal','get_live_exporters')
    def get_old_exporters(self):
        keys = self.old_exporters.keys()
        keys.sort()
        keys.reverse()
        return keys

    security.declareProtected('Manage Portal','get_exporter_info')
    def get_exporter_info(self,xptr_id=None,REQUEST=None,RESPOSE=None):
        """return a dict with exporter info"""
        xptr_id = xptr_id or REQUEST.form.get('xptr_id','')
        if xptr_id in live_exporters or xptr_id in self.old_exporters:
            return self._xptr_info(xptr_id)

    security.declareProtected('Manage Portal','_xptr_info')
    def _xptr_info(self,xptr_id,xml=False):
        result = {}
        can_I_run = True
        try:
            xptr = live_exporters[xptr_id]
        except KeyError:
            xptr = self.old_exporters[xptr_id]
            can_I_run = False
        result['state'] = xptr.running and 'running' or 'stopped'
        result['counter'] = xptr.counter
        if xptr.approx_ammount:
            if xptr.running:
                result['progress'] = '%.1f %%' % min(99.0,(xptr.counter*100.0/xptr.approx_ammount))
            elif xptr.report == 'done':
                result['progress'] = '100 %'
        else:
            result['progress'] = '--'
        result['start_from'] = xptr.from_path
        result['report'] = xptr.report or '--'
        result['started'] = xptr.started or '--'
        result['finished'] = xptr.finished or '--'
        result['runnable'] = can_I_run and xptr.started is None
        result['time'] = elapsed_time(xptr.started,xptr.finished)
        if xml:
            xml_text = '<xptr id="%s" ' % xptr_id
            for k,v in result.items():
                xml_text += ' %s="%s"' %(k,v)
            xml_text += '></xptr>'
            return xml_text
        return result

    security.declareProtected('Manage Portal','get_tar_gz')
    def get_tar_gz(self,REQUEST=None,RESPOSE=None):
        """return the tarball of the given exporter id"""
        xptr_id = REQUEST.form.get('xptr_id','')
        try:
            xptr = self.old_exporters[xptr_id]
        except KeyError:
            return 'Invalid exported id %s' % xptr_id
        exported_path = xptr._base_path

        tar_gz_id = '%s.tar.gz' % exported_path
        if not os.path.exists(tar_gz_id):
            if not os.path.exists(exported_path):
                return 'There are no exported results for id %s' % xptr_id
            tar = tarfile.open(tar_gz_id, "w:gz")
            os.chdir(os.path.dirname(exported_path))
            tar.add(os.path.basename(exported_path))
            tar.close()
        REQUEST.response.setHeader('content-type','application/x-tar')
        REQUEST.response.setHeader('content-disposition',
                            'inline; filename=%s' % os.path.basename(tar_gz_id));
        return file(tar_gz_id).read()

    security.declareProtected('Manage Portal','clear_old_exporters')
    def clear_old_exporters(self, REQUEST=None, RESPONSE=None):
        """Forgets all previous xptr information objects, but not the exported
           results on file system"""
        self.old_exporters = PersistentMapping()
        return RESPONSE.redirect(getToolByName(self,'portal_url').getPortalObject().absolute_url()+'/conf_sql_exporter')

InitializeClass(SQLExporterTool)
