import os
from datetime import datetime
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.utils import base_hasattr
from Products.SQLExporter import mongerorm
from Products.Archetypes.Field import FileField, TextField
reload(mongerorm)

from OFS.SimpleItem import SimpleItem

dublin_core_accessors = {
    'contributors':'Contributors',
    'created': 'CreationDate',
    'creators':'Creators',
    'description':'Description',
    'effective': 'EffectiveDate',
    'expires': 'ExpirationDate',
    'modified': 'ModificationDate',
    'publisher': 'Publisher',
    'subject': 'Subject',
    'title': 'Title',
    'rights': 'Rights',
    'language': 'Language',
    'format': 'Format'
}

db_connections = {}

reference_field_types = ['Products.Archetypes.Field.ReferenceField',
                         'Products.OrderableReferenceField._field.OrderableReferenceField']

strip_url_path_types = []
try:
    from Products.Relations.ruleset import Ruleset
except ImportError:
    pass
else:
    strip_url_path_types.append(Ruleset)
strip_url_path_types = tuple(strip_url_path_types)

attachment_file_fields = (FileField,)

def attachable_field(field):
    return isinstance(field,attachment_file_fields) and\
           not isinstance(field,TextField)

def serialize(o):
    return repr(o)

class Exporter(SimpleItem):
    def __init__(self,id,from_path,database_path,
                 excluded_ids,excluded_meta_types):
        self.from_path = from_path
        self.approx_ammount = 0
        self.id = id
        self.setup_folders(database_path)
        self._files_counter = 0
        self.excluded_ids = excluded_ids[:]
        self.excluded_meta_types = excluded_meta_types[:]
        self.running = False
        self.counter = 0
        self.report = ""
        self.finished = None
        self.started = None

    def get_monger(self):
        return db_connections.get(self.id)
    def set_monger(self,value):
        db_connections[self.id] = value
    monger = property(get_monger,set_monger)

    def get_next_file_counter(self):
        self._files_counter += 1
        result = self._files_counter
        return result

    def abs_url_path_stripped(self,item):
        """returns item absolute_url_path minus the from_path"""
        item_path = item.absolute_url_path()

        if item_path.startswith(self.from_path):
            item_path = item_path[len(self.from_path):]

        return item_path

    def setup_folders(self,database_path):
        self._base_path = os.path.join(database_path,'importtool_exports_%s' % self.id)
        if not os.path.exists(self._base_path):
            os.makedirs(self._base_path)
        elif not os.path.isdir(database_path):
            raise IOError('The selected database path %s exists but is not a folder' % database_path)

        if not os.path.exists(os.path.join(self._base_path,'__init__.py')):
            open(os.path.join(self._base_path,'__init__.py'),'w')

        self.files_path = os.path.join(self._base_path,'external_files')
        if not os.path.exists(self.files_path):
            os.makedirs(self.files_path)

    def download_and_attach(self,field,plone_item):
        value = getattr(plone_item,field.accessor)()
        if not value:
            return None
        else:
            expected_size = value.getSize()
            if base_hasattr(value,'data') and len(str(value.data)) == expected_size:
                file_data = str(value.data)
            #elif value is None or value.__class__.__name__ == 'File':
            #    file_data = str(value)
            elif len(str(value)) == expected_size:
                file_data = str(value)
            else:
                try:
                    file_data = value.download()
                except AttributeError:
                    file_data = field.download(plone_item)
            if len(file_data) != expected_size:
                print 'I know there are problems with the attached file of %s' % plone_item.absolute_url_path()
        files_counter = self.get_next_file_counter()
        path = os.path.join(self.files_path,str(files_counter))
        f = file(path,'w')
        f.write(file_data)
        f.close()
        return files_counter

    def get_dublin_core(self,obj):
        d = {}
        for attr,accessor in dublin_core_accessors.items():
            try:
                d[attr] = serialize(getattr(obj,accessor)())
            except AttributeError:
                if obj.meta_type == 'Subscriber' and attr == 'creators':
                    d[attr] = serialize(obj.creators)
        return d

    def get_schema_fields(self,obj):
        d = {}
        try:
            fields = obj.schema.fields()
        except AttributeError:
            return {}
        for field in fields:
            if attachable_field(field):
                d[field.__name__] = serialize((self.download_and_attach(field,obj),field.getFilename(obj)))
                continue
            try:
                value = getattr(obj,field.accessor)()
            except AttributeError, error:
                if field.__name__ == 'relatedItems' and 'order' in str(error):
                    # strangely, calling to getRaw first,
                    # make the AttributeError dissapear
                    raw = obj.getRawRelatedItems()
                    value = getattr(obj,field.accessor)()
                else:
                    raise
            except KeyError, error:
                if field.__name__ == 'body_pt' and 'fields' in str(error):
                    continue
                else:
                    raise
            except TypeError, error:
                if field.accessor is None and hasattr(obj,field.__name__):
                    value = getattr(obj,field.__name__)
                    if callable(value):
                        value = value()
                else:
                    raise

            if field.getType() in reference_field_types:
                if value is None:
                    pass
                elif strip_url_path_types and isinstance(value,strip_url_path_types):
                    value = self.abs_url_path_stripped(value)
                else:
                    try:
                        iterable = iter(value)
                    except TypeError:
                        # value is not iterable
                        value = self.abs_url_path_stripped(value)
                    else:
                        try:
                            value = [self.abs_url_path_stripped(item) for item in iterable]
                        except Exception, e:
                            print 'Error occurred:',e

            d[field.__name__] = serialize(value)

        if getattr(obj,'portal_type','') == 'Subscriber':
            d['active'] = serialize(obj.active)
            d['format'] = serialize(obj.format)
        if getattr(obj,'portal_type','') == 'Newsletter':
            d['dateEmitted'] = serialize(obj.dateEmitted)
        return d

    def get_properties(self,obj):
        d = {}
        try:
            d['_propertyMap'] = serialize(obj._propertyMap())
            d['propertyItems'] = serialize(obj.propertyItems())
        except (AttributeError, TypeError):
            return {}
        return d

    def workflow_info(self,obj):
        d = {}
        try:
            d['workflow_state'] = serialize(self.pworkflow.getInfoFor(obj,'review_state'))
        except Exception, e:
            d['workflow_state'] = serialize(None)
        try:
            d['workflow_id'] = serialize(self.pworkflow.getChainFor(obj))
        except Exception, e:
            d['workflow_id'] = serialize(None)
        return d

    def get_basic_attrs(self,obj):
        d = {}
        d['id'] = serialize(obj.getId())
        for key in ['portal_type','meta_type']:
            d[key] = serialize(getattr(obj,key,None))
        d['owner'] = serialize(obj.getOwner())
        return d

    def attrs_dict(self,obj):
        d = {}
        for subset_method in [self.get_dublin_core,self.get_schema_fields,
                              self.get_properties,self.get_basic_attrs,
                              self.workflow_info]:
            try:
                d.update(subset_method(obj))
            except:
                print 'creating the attrs_dict, ', subset_method, 'failed'
                raise
        return d

    def run(self,folder):
        # we update the from_path value:
        self.from_path = folder.absolute_url_path()
        self.pworkflow = getToolByName(folder,'portal_workflow')
        self.approx_ammount = len(getToolByName(folder,'portal_catalog')(path=self.from_path))
        self.counter = 0
        self.running = True
        self.started = datetime.now()
        try:
            self.monger = mongerorm.SQLMonger(os.path.join(self._base_path,'database.sqlite'))
            info_file = file(os.path.join(self._base_path,'info.py'),'w')
            info_file.write('from_path = "%s"' % self.from_path)
            info_file.close()
            self.folder_dig(folder)
        except Exception, e:
            msg = 'We found a problem and the process was interrupted: %s' % e
        else:
            msg = 'done'
        self.finished = datetime.now()
        self.running = False
        self.report = msg
        return msg

    def folder_dig(self,folder):
        print 'Digging folder', folder
        for obj in folder.listFolderContents():
            self.increment_counter()
            if (getattr(obj,'id','') in self.excluded_ids or
                getattr(obj,'meta_type','') in self.excluded_meta_types):
                continue
            d = self.attrs_dict(obj)
            self.monger.store(self.abs_url_path_stripped(obj),
                         self.abs_url_path_stripped(folder),
                         attrs=d)
            if obj.isPrincipiaFolderish:
                self.folder_dig(obj)

    def increment_counter(self):
        self.counter += 1
